#!/usr/bin/env python3
# dac8532.py
#
# Copyright (c) 2020-2022 Spring City Solutions LLC and other contributors
# Licensed under the Apache License, Version 2.0
#
# Need to have the rpi.gpio and spidev python libraries installed
#
# For development, need to install the pytest system.
# sudo apt-get install python3-pytest
# See also tests/dac8532_test.py
# Running 'python -m pytest' should result in no errors
#
# For development, need to install the flake8 system.
# sudo apt-get install python3-flake8
# Running flake8 should report no serious problems,
# with E501 being ignored and excluding tests/
#
# Hardware Links:
#
# Board Manufacturer Store:
# https://www.waveshare.com/High-Precision-AD-DA-Board.htm
#
# Board Manufacturer Wiki:
# https://www.waveshare.com/wiki/High-Precision_AD/DA_Board
#
# Links for ADS1256 chip:
# https://www.ti.com/product/ADS1256
# https://www.ti.com/lit/gpn/ads1256
# Book "Fundamentals of Precision ADC Noise Analysis" https://www.ti.com/lit/pdf/slyy192
# Journal Series "How delta-sigma ADCs work" https://www.ti.com/lit/pdf/slyt423 https://www.ti.com/lit/pdf/slyt438
# App Note "Digital Filter Types in Delta-Sigma ADCs" https://www.ti.com/lit/pdf/sbaa230
#
# Link for LM285 chip, connected to the ADS1256:
# https://www.ti.com/product/LM285-2.5-N
# https://www.ti.com/lit/gpn/lm285-2.5-n
#
# Link for DAC8532 chip:
# https://www.ti.com/product/DAC8532
# https://www.ti.com/lit/gpn/dac8532
#
# Spring City Solutions Links:
#
# Spring City Solutions Node-RED project page:
# https://www.springcitysolutions.com/nodered
#
# Spring City Solutions page for this node:
# https://www.springcitysolutions.com/nodered-waveshare-adda-shield
#
# Spring City Solutions Node-RED project email:
# nodered@springcitysolutions.com
#
# Spring City Solutions Gitlab for this node:
# https://gitlab.com/SpringCitySolutionsLLC/waveshare-shield-adda-11010
#
# Patreon Page:
# https://www.patreon.com/springcitysolutions_nodered
#
# Software Links:
#
# Spring City Solutions Gitlab for this node:
# https://gitlab.com/SpringCitySolutionsLLC/waveshare-shield-adda-11010
#
# Doxygen docs for this node autogenerated by Gitlab CI/CD:
# https://springcitysolutionsllc.gitlab.io/waveshare-shield-adda-11010/index.html
#
# npmjs for this node:
# https://npmjs.org/package/node-red-contrib-waveshare-shield-adda-11010
#
# Node-RED flows for this node:
# https://flows.nodered.org/node/node-red-contrib-waveshare-shield-adda-11010
#
# Documentation Links:
#
# Gitlab wiki for this node (the master copy of list of links is here):
# https://gitlab.com/SpringCitySolutionsLLC/waveshare-shield-adda-11010/-/wikis/home
#
# Waveshare's libraries for this hardware:
# https://github.com/waveshare/High-Precision-AD-DA-Board
#
# Youtube video "How to set up":
# TODO
#
# Youtube video "How to use":
# TODO
#
# Youtube video "Testing Results":
# TODO
#
# Usage:
#
# dac8532.py "port A or B" "integer value 0 to 65535"
#
# Usage:
#
# Combine any of the following options:
#
# dac8532.py -t
# Returns a JSON of the register data for use with pytest or other testing
#
# ads1015.py --port "default A"
# Selects output port from one of the following options:
# A, B
#
# ads1015.py --output "default 32767"
# Sets the output to unsigned 16 bit integer from 0 to 65535

import argparse
import json
import sys

import RPi.GPIO as GPIO
import spidev

cs = 23

register_data = dict([('control', 0x00), ('data_h', 0x00), ('data_l', 0x00)])


# Functions that do things.


def output_json():
    """Prints a JSON of the register_data for development and testing."""
    print(json.dumps(register_data, sort_keys=True, indent=4))


def parse_args(raw_args):
    """Parses the args into ready to write register_data."""
    # First parse the CLI.
    parser = argparse.ArgumentParser()
    parser.add_argument("--port", type=str, default="A",
                        help="Output Port: A, B default A")
    parser.add_argument("--output", type=str, default="32767",
                        help="Output Voltage: 0 to 65535 default 32767")
    parser.add_argument("-t", "--test", action="store_true",
                        help="Print JSON of register_data")
    args = parser.parse_args(raw_args)

    # Process the control register
    register_data['control'] = 0x30
    if (args.port == "A"):
        register_data['control'] = 0x30
    if (args.port == "B"):
        register_data['control'] = 0x34

    # Process the data registers
    register_data['data_h'] = 0x00
    register_data['data_l'] = 0x00
    raw_output = int(args.output)
    if (raw_output < 0):
        raw_output = 0
    if (raw_output > 65535):
        raw_output = 65535
    register_data['data_h'] = raw_output >> 8
    register_data['data_l'] = raw_output & 0xFF

    # Optionally output register state for development purposes
    if args.test:
        output_json()
        exit()


# Functions that read or write device registers


def write():
    """Writes to the DAC using SPI"""

    GPIO.setmode(GPIO.BCM)
    GPIO.setwarnings(False)
    GPIO.setup(cs, GPIO.OUT)
    GPIO.output(cs, 1)

    spi = spidev.SpiDev()
    spi.open(0, 0)
    spi.max_speed_hz = 20000
    spi.mode = 0b01

    GPIO.output(cs, 0)
    spi.writebytes([register_data['control'], register_data['data_h'], register_data['data_l']])
    GPIO.output(cs, 1)


# Convert CLI options to register_data
parse_args(sys.argv[1:])

# Write to the DAC
write()
