#!/usr/bin/env python3
# gpio.py
#
# Copyright (c) 2020-2022 Spring City Solutions LLC and other contributors
# Licensed under the Apache License, Version 2.0
#
# Need to have the rpi.gpio python library installed
#
# For development, need to install the pytest system.
# sudo apt-get install python3-pytest
# See also tests/gpio_test.py
# Running 'python -m pytest' should result in no errors
#
# For development, need to install the flake8 system.
# sudo apt-get install python3-flake8
# Running flake8 should report no serious problems,
# with E501 being ignored and excluding tests/
#
# Some notes on wiring:
# Labeled header P22 is connected to physical pin 31 connected to BCM port 6
# Labeled header P23 is connected to physical pin 33 connected to BCM port 13
# Labeled header P24 is connected to physical pin 35 connected to BCM port 19
# Labeled header P25 is connected to physical pin 37 connected to BCM port 26
#
# Hardware Links:
#
# Board Manufacturer Store:
# https://www.waveshare.com/High-Precision-AD-DA-Board.htm
#
# Board Manufacturer Wiki:
# https://www.waveshare.com/wiki/High-Precision_AD/DA_Board
#
# Links for ADS1256 chip:
# https://www.ti.com/product/ADS1256
# https://www.ti.com/lit/gpn/ads1256
# Book "Fundamentals of Precision ADC Noise Analysis" https://www.ti.com/lit/pdf/slyy192
# Journal Series "How delta-sigma ADCs work" https://www.ti.com/lit/pdf/slyt423 https://www.ti.com/lit/pdf/slyt438
# App Note "Digital Filter Types in Delta-Sigma ADCs" https://www.ti.com/lit/pdf/sbaa230
#
# Link for LM285 chip, connected to the ADS1256:
# https://www.ti.com/product/LM285-2.5-N
# https://www.ti.com/lit/gpn/lm285-2.5-n
#
# Link for DAC8532 chip:
# https://www.ti.com/product/DAC8532
# https://www.ti.com/lit/gpn/dac8532
#
# Spring City Solutions Links:
#
# Spring City Solutions Node-RED project page:
# https://www.springcitysolutions.com/nodered
#
# Spring City Solutions page for this node:
# https://www.springcitysolutions.com/nodered-waveshare-adda-shield
#
# Spring City Solutions Node-RED project email:
# nodered@springcitysolutions.com
#
# Spring City Solutions Gitlab for this node:
# https://gitlab.com/SpringCitySolutionsLLC/waveshare-shield-adda-11010
#
# Patreon Page:
# https://www.patreon.com/springcitysolutions_nodered
#
# Software Links:
#
# Spring City Solutions Gitlab for this node:
# https://gitlab.com/SpringCitySolutionsLLC/waveshare-shield-adda-11010
#
# Doxygen docs for this node autogenerated by Gitlab CI/CD:
# https://springcitysolutionsllc.gitlab.io/waveshare-shield-adda-11010/index.html
#
# npmjs for this node:
# https://npmjs.org/package/node-red-contrib-waveshare-shield-adda-11010
#
# Node-RED flows for this node:
# https://flows.nodered.org/node/node-red-contrib-waveshare-shield-adda-11010
#
# Documentation Links:
#
# Gitlab wiki for this node (the master copy of list of links is here):
# https://gitlab.com/SpringCitySolutionsLLC/waveshare-shield-adda-11010/-/wikis/home
#
# Waveshare's libraries for this hardware:
# https://github.com/waveshare/High-Precision-AD-DA-Board
#
# Youtube video "How to set up":
# TODO
#
# Youtube video "How to use":
# TODO
#
# Youtube video "Testing Results":
# TODO
#
# Usage:
#
# Combine any of the following options:
#
# gpio.py -t
# Returns a JSON of the register data for use with pytest
#
# gpio.py --pin "default P22"
# Sets port to one of the following options:
# P22, P23, P24, P25
#
# gpio.py --output
# No output option means input mode; prints a 0 or 1 depending on port input
# Specifying 0 or 1 as an output puts the port into output mode and sets the port to 0 or 1
#
# Note, if you have an open circuit unconnected port, and
# output a 1, then run a input a few times, the physical wiring
# has some capacitance and will input a 1 for the first
# read, but later reads will read as 0.
# It probably doesn't matter in normal operation, but it looks weird.
#

import argparse
import json
import sys

import RPi.GPIO as GPIO

register_data = dict([('pin', 0), ('output', 0)])


# Functions that do things.


def output_json():
    """Prints a JSON of the register_data for development and testing."""
    print(json.dumps(register_data, sort_keys=True, indent=4))


def parse_args(raw_args):
    """Parses the args into ready to write register_data."""
    # First parse the CLI.
    parser = argparse.ArgumentParser()
    parser.add_argument("--pin", type=str, default="P22",
                        help="Pin: P22, P23, P24, P25 default P22")
    parser.add_argument("--output", type=str, default="-1",
                        help="Providing this puts the pin in output mode instead of input mode.  Output: 0, 1")
    parser.add_argument("-t", "--test", action="store_true",
                        help="Print JSON of register_data")
    args = parser.parse_args(raw_args)

    # Process pin selection
    register_data['pin'] = 6
    if (args.pin == 'P22'):
        register_data['pin'] = 6
    if (args.pin == 'P23'):
        register_data['pin'] = 13
    if (args.pin == 'P24'):
        register_data['pin'] = 19
    if (args.pin == 'P25'):
        register_data['pin'] = 26

    # Process output, which also selects input or output mode
    register_data['output'] = -1
    if (args.output == '0'):
        register_data['output'] = 0
    if (args.output == '1'):
        register_data['output'] = 1

    # Optionally output register state for development purposes.
    if args.test:
        output_json()
        exit()


# Functions that read or write device


def config():
    """Configure the port mode"""
    GPIO.setmode(GPIO.BCM)
    GPIO.setwarnings(False)

    if (register_data['output'] == -1):
        GPIO.setup(register_data['pin'], GPIO.IN)
        print(GPIO.input(register_data['pin']))

    if (register_data['output'] != -1):
        GPIO.setup(register_data['pin'], GPIO.OUT)
        GPIO.output(register_data['pin'], register_data['output'])


# Convert CLI options to register_data.
parse_args(sys.argv[1:])

# Configure the pin
config()
